[From nobody Wed Nov  7 16:41:36 2007
From: Karsten Wiese &lt;fzu@wemgehoertderstaat.de&gt;
To: linux-usb-devel@lists.sourceforge.net
Subject: [PATCH] ehci-hcd: complete iso urbs ASAP for number_of_packets != (n
	* 8) also
Date: Wed, 7 Nov 2007 16:16:34 +0100
User-Agent: KMail/1.9.6 (enterprise 0.20070907.709405)
MIME-Version: 1.0
Content-Type: text/plain;
  charset=&quot;us-ascii&quot;
Content-Transfer-Encoding: 7bit
Content-Disposition: inline
Message-Id: &lt;200711071616.34194.fzu@wemgehoertderstaat.de&gt;


Upstream (as of 2.6.23) ehci-hcd only completes iso urbs, if the last frame
(= 8 uframes) they overlapped with has elapsed. That can be as late as when
the following urb emits its interrupt on completion.
soundcard drivers tend to work around by only transfering iso urbs with
number_of_packets = (n * 8).
Patch lets iso urbs complete asap by scanning itd always up to the elapsed
uframe. An itd's last scanned uframe is stored in the new struct ehci_itd
member uframe_scanned.
Itds stay kept in the hcd's schedule until the complete frame they cover
has elapsed, like without patch.
To make this possible, ehci-&gt;periodic_sched is changed based on the number of
active itds instead of the number of active iso urbs.

Signed-off-by: Karsten Wiese &lt;fzu@wemgehoertderstaat.de&gt;
---
 drivers/usb/host/ehci-sched.c |  108 ++++++++++++++++++++---------------------
 drivers/usb/host/ehci.h       |    1 +
 2 files changed, 54 insertions(+), 55 deletions(-)

diff --git a/drivers/usb/host/ehci-sched.c b/drivers/usb/host/ehci-sched.c
index e682f23..c456b16 100644
--- a/drivers/usb/host/ehci-sched.c
+++ b/drivers/usb/host/ehci-sched.c
@@ -1481,6 +1481,8 @@ itd_link (struct ehci_hcd *ehci, unsigned frame, struct ehci_itd *itd)
 	itd-&gt;frame = frame;
 	wmb ();
 	ehci-&gt;periodic[frame] = cpu_to_hc32(ehci, itd-&gt;itd_dma | Q_TYPE_ITD);
+	if (unlikely (!ehci-&gt;periodic_sched++))
+		enable_periodic (ehci);
 }
 
 /* fit urb's itds into the selected schedule slot; activate as needed */
@@ -1553,30 +1555,30 @@ itd_link_urb (
 	urb-&gt;hcpriv = NULL;
 
 	timer_action (ehci, TIMER_IO_WATCHDOG);
-	if (unlikely (!ehci-&gt;periodic_sched++))
-		return enable_periodic (ehci);
 	return 0;
 }
 
 #define	ISO_ERRS (EHCI_ISOC_BUF_ERR | EHCI_ISOC_BABBLE | EHCI_ISOC_XACTERR)
 
-static unsigned
-itd_complete (
-	struct ehci_hcd	*ehci,
-	struct ehci_itd	*itd
-) {
+static bool
+itd_scan (
+	struct ehci_hcd *ehci,
+	struct ehci_itd	*itd,
+	unsigned uframe_after
+)
+{
 	struct urb				*urb = itd-&gt;urb;
 	struct usb_iso_packet_descriptor	*desc;
 	u32					t;
 	unsigned				uframe;
 	int					urb_index = -1;
 	struct ehci_iso_stream			*stream = itd-&gt;stream;
-	struct usb_device			*dev;
 
 	/* for each uframe with a packet */
-	for (uframe = 0; uframe &lt; 8; uframe++) {
+	for (uframe = itd-&gt;uframe_scanned; uframe &lt; uframe_after; uframe++) {
 		if (likely (itd-&gt;index[uframe] == -1))
 			continue;
+
 		urb_index = itd-&gt;index[uframe];
 		desc = &amp;urb-&gt;iso_frame_desc [urb_index];
 
@@ -1603,45 +1605,53 @@ itd_complete (
 			desc-&gt;status = 0;
 			desc-&gt;actual_length = EHCI_ITD_LENGTH (t);
 		}
+
+		/* handle completion now? */
+		if (urb_index + 1 == urb-&gt;number_of_packets) {
+			struct usb_device *dev;
+
+			usb_put_urb (urb);
+			/* ASSERT: it's really the last itd for this urb
+			   struct ehci_itd	*stritd;
+			   list_for_each_entry (stritd, &amp;stream-&gt;td_list, itd_list)
+			   BUG_ON (stritd != itd &amp;&amp; stritd-&gt;urb == urb);
+			*/
+			itd-&gt;urb = NULL;
+
+			/* give urb back to the driver ... can be out-of-order */
+			dev = urb-&gt;dev;
+			ehci_urb_done (ehci, urb);
+			urb = NULL;
+
+			ehci_to_hcd(ehci)-&gt;self.bandwidth_isoc_reqs--;
+
+			if (unlikely (list_empty (&amp;stream-&gt;td_list))) {
+				ehci_to_hcd(ehci)-&gt;self.bandwidth_allocated
+					-= stream-&gt;bandwidth;
+				ehci_vdbg (ehci,
+					   &quot;deschedule devp %s ep%d%s-iso\n&quot;,
+					   dev-&gt;devpath, stream-&gt;bEndpointAddress &amp; 0x0f,
+					   (stream-&gt;bEndpointAddress &amp; USB_DIR_IN) ? &quot;in&quot; : &quot;out&quot;);
+			}
+			iso_stream_put (ehci, stream);
+		}
 	}
 
-	usb_put_urb (urb);
-	itd-&gt;urb = NULL;
+	itd-&gt;uframe_scanned = uframe_after;
+	if (uframe_after &lt; 8)
+		return false;
+
 	itd-&gt;stream = NULL;
+	itd-&gt;uframe_scanned = 0;
 	list_move (&amp;itd-&gt;itd_list, &amp;stream-&gt;free_list);
 	iso_stream_put (ehci, stream);
 
-	/* handle completion now? */
-	if (likely ((urb_index + 1) != urb-&gt;number_of_packets))
-		return 0;
-
-	/* ASSERT: it's really the last itd for this urb
-	list_for_each_entry (itd, &amp;stream-&gt;td_list, itd_list)
-		BUG_ON (itd-&gt;urb == urb);
-	 */
-
-	/* give urb back to the driver ... can be out-of-order */
-	dev = urb-&gt;dev;
-	ehci_urb_done (ehci, urb);
-	urb = NULL;
-
 	/* defer stopping schedule; completion can submit */
 	ehci-&gt;periodic_sched--;
 	if (unlikely (!ehci-&gt;periodic_sched))
 		(void) disable_periodic (ehci);
-	ehci_to_hcd(ehci)-&gt;self.bandwidth_isoc_reqs--;
 
-	if (unlikely (list_empty (&amp;stream-&gt;td_list))) {
-		ehci_to_hcd(ehci)-&gt;self.bandwidth_allocated
-				-= stream-&gt;bandwidth;
-		ehci_vdbg (ehci,
-			&quot;deschedule devp %s ep%d%s-iso\n&quot;,
-			dev-&gt;devpath, stream-&gt;bEndpointAddress &amp; 0x0f,
-			(stream-&gt;bEndpointAddress &amp; USB_DIR_IN) ? &quot;in&quot; : &quot;out&quot;);
-	}
-	iso_stream_put (ehci, stream);
-
-	return 1;
+	return true;
 }
 
 /*-------------------------------------------------------------------------*/
@@ -2156,30 +2166,18 @@ restart:
 				q = q.fstn-&gt;fstn_next;
 				break;
 			case Q_TYPE_ITD:
-				/* skip itds for later in the frame */
 				rmb ();
-				for (uf = live ? uframes : 8; uf &lt; 8; uf++) {
-					if (0 == (q.itd-&gt;hw_transaction [uf]
-							&amp; ITD_ACTIVE(ehci)))
-						continue;
+				modified = itd_scan(ehci, q.itd, uf = live ? uframes : 8);
+				if (!modified) {
 					q_p = &amp;q.itd-&gt;itd_next;
 					hw_p = &amp;q.itd-&gt;hw_next;
-					type = Q_NEXT_TYPE(ehci,
-							q.itd-&gt;hw_next);
-					q = *q_p;
-					break;
+				} else {
+					*q_p = q.itd-&gt;itd_next;
+					*hw_p = q.itd-&gt;hw_next;
 				}
-				if (uf != 8)
-					break;
-
-				/* this one's ready ... HC won't cache the
-				 * pointer for much longer, if at all.
-				 */
-				*q_p = q.itd-&gt;itd_next;
-				*hw_p = q.itd-&gt;hw_next;
 				type = Q_NEXT_TYPE(ehci, q.itd-&gt;hw_next);
+
 				wmb();
-				modified = itd_complete (ehci, q.itd);
 				q = *q_p;
 				break;
 			case Q_TYPE_SITD:
diff --git a/drivers/usb/host/ehci.h b/drivers/usb/host/ehci.h
index 951d69f..df35964 100644
--- a/drivers/usb/host/ehci.h
+++ b/drivers/usb/host/ehci.h
@@ -581,6 +581,7 @@ struct ehci_itd {
 	struct urb		*urb;
 	struct ehci_iso_stream	*stream;	/* endpoint's queue */
 	struct list_head	itd_list;	/* list of stream's itds */
+	unsigned		uframe_scanned;
 
 	/* any/all hw_transactions here may be used by that urb */
 	unsigned		frame;		/* where scheduled */
-- 
1.5.3.3

]